# coding: utf-8
# codeskyblue 2020/06/03
#

__all__ = [
    'Color', 'AFC_MAGIC', 'AFC', 'AFCStatus', 'LOCKDOWN_PORT', 'PROGRAM_NAME',
    'SYSMON_PROC_ATTRS', 'SYSMON_SYS_ATTRS', 'MODELS', 'LockdownService',
    "UsbmuxReplyCode", "InstrumentsService", "LOG", "StatResult"
]

from dataclasses import dataclass
import datetime
import enum


class LOG(str, enum.Enum):
    main = "tidevice"
    xctest = "tidevice.xctest"


class Color(enum.Enum):
    END = '\033[0m'
    BOLD = '\033[1m'

    # Foreground
    BLACK = '\33[30m'
    RED = '\33[31m'
    GREEN = '\33[32m'
    YELLOW = '\33[33m'
    BLUE = '\33[34m'
    VIOLET = '\33[35m'
    BEIGE = '\33[36m'
    WHITE = '\33[37m'

    # Background
    BG_GREEN = '\033[42m'

    @staticmethod
    def wrap_text(text, *colors):
        prefix = ''.join([c.value for c in colors])
        return prefix + text + Color.END.value


SYSMON_PROC_ATTRS = [
    "memVirtualSize",  # vss
    "cpuUsage",
    "ctxSwitch",  # the number of context switches by process each second
    "intWakeups",  # the number of threads wakeups by process each second
    "physFootprint",  # real memory (物理内存)
    "memResidentSize",  # rss
    "memAnon",  # anonymous memory
    "pid"
]

SYSMON_SYS_ATTRS = [
    "vmExtPageCount", "vmFreeCount", "vmPurgeableCount", "vmSpeculativeCount",
    "physMemSize"
]

LOCKDOWN_PORT = 62078
PROGRAM_NAME = "tidevice"

AFC_MAGIC = b"CFA6LPAA"


class AFC(enum.IntEnum):
    OP_INVALID = 0x00000000
    OP_STATUS = 0x00000001
    OP_DATA = 0x00000002  # Data
    OP_READ_DIR = 0x00000003  # ReadDir
    OP_READ_FILE = 0x00000004  # ReadFile
    OP_WRITE_FILE = 0x00000005  # WriteFile
    OP_WRITE_PART = 0x00000006  # WritePart
    OP_TRUNCATE = 0x00000007  # TruncateFile
    OP_REMOVE_PATH = 0x00000008  # RemovePath
    OP_MAKE_DIR = 0x00000009  # MakeDir
    OP_GET_FILE_INFO = 0x0000000a  # GetFileInfo
    OP_GET_DEVINFO = 0x0000000b  # GetDeviceInfo
    OP_WRITE_FILE_ATOM = 0x0000000c  # WriteFileAtomic (tmp file+rename)
    OP_FILE_OPEN = 0x0000000d  # FileRefOpen
    OP_FILE_OPEN_RES = 0x0000000e  # FileRefOpenResult
    OP_READ = 0x0000000f  # FileRefRead
    OP_WRITE = 0x00000010  # FileRefWrite
    OP_FILE_SEEK = 0x00000011  # FileRefSeek
    OP_FILE_TELL = 0x00000012  # FileRefTell
    OP_FILE_TELL_RES = 0x00000013  # FileRefTellResult
    OP_FILE_CLOSE = 0x00000014  # FileRefClose
    OP_FILE_SET_SIZE = 0x00000015  # FileRefSetFileSize (ftruncate)
    OP_GET_CON_INFO = 0x00000016  # GetConnectionInfo
    OP_SET_CON_OPTIONS = 0x00000017  # SetConnectionOptions
    OP_RENAME_PATH = 0x00000018  # RenamePath
    OP_SET_FS_BS = 0x00000019  # SetFSBlockSize (0x800000)
    OP_SET_SOCKET_BS = 0x0000001A  # SetSocketBlockSize (0x800000)
    OP_FILE_LOCK = 0x0000001B  # FileRefLock
    OP_MAKE_LINK = 0x0000001C  # MakeLink
    OP_SET_FILE_TIME = 0x0000001E  # set st_mtime
    OP_GET_FILE_HASH_RANGE = 0x0000001F  # GetFileHashWithRange

    O_RDONLY = 0x00000001  #/**< r   O_RDONLY
    O_RW = 0x00000002  #/**< r+  O_RDWR   | O_CREAT
    O_WRONLY = 0x00000003  #/**< w   O_WRONLY | O_CREAT  | O_TRUNC
    O_WR = 0x00000004  #/**< w+  O_RDWR   | O_CREAT  | O_TRUNC
    O_APPEND = 0x00000005  #/**< a   O_WRONLY | O_APPEND | O_CREAT
    O_RDAPPEND = 0x00000006  #/**< a+  O_RDWR   | O_APPEND | O_CREAT

    HARDLINK = 1
    SYMLINK = 2

    LOCK_SH = 1 | 4  #/**< shared lock
    LOCK_EX = 2 | 4  #/**< exclusive lock
    LOCK_UN = 8 | 4  #/**< unlock

    # #// Status
    # ST_SUCCESS                = 0
    # ST_UNKNOWN_ERROR          = 1
    # ST_OP_HEADER_INVALID      = 2
    # ST_NO_RESOURCES           = 3
    # ST_READ_ERROR             = 4
    # ST_WRITE_ERROR            = 5
    # ST_UNKNOWN_PACKET_TYPE    = 6
    # ST_INVALID_ARG            = 7
    # ST_OBJECT_NOT_FOUND       = 8
    # ST_OBJECT_IS_DIR          = 9
    # ST_PERM_DENIED            =10
    # ST_SERVICE_NOT_CONNECTED  =11
    # ST_OP_TIMEOUT             =12
    # ST_TOO_MUCH_DATA          =13
    # ST_END_OF_DATA            =14
    # ST_OP_NOT_SUPPORTED       =15
    # ST_OBJECT_EXISTS          =16
    # ST_OBJECT_BUSY            =17
    # ST_NO_SPACE_LEFT          =18
    # ST_OP_WOULD_BLOCK         =19
    # ST_IO_ERROR               =20
    # ST_OP_INTERRUPTED         =21
    # ST_OP_IN_PROGRESS         =22
    # ST_INTERNAL_ERROR         =23

    # ST_MUX_ERROR              =30
    # ST_NO_MEM                 =31
    # ST_NOT_ENOUGH_DATA        =32
    # ST_DIR_NOT_EMPTY          =33


@enum.unique
class AFCStatus(enum.IntEnum):
    SUCCESS = 0
    UNKNOWN_ERROR = 1
    OP_HEADER_INVALID = 2
    NO_RESOURCES = 3
    READ_ERROR = 4
    WRITE_ERROR = 5
    UNKNOWN_PACKET_TYPE = 6
    INVALID_ARG = 7
    OBJECT_NOT_FOUND = 8
    OBJECT_IS_DIR = 9
    PERM_DENIED = 10
    SERVICE_NOT_CONNECTED = 11
    OP_TIMEOUT = 12
    TOO_MUCH_DATA = 13
    END_OF_DATA = 14
    OP_NOT_SUPPORTED = 15
    OBJECT_EXISTS = 16
    OBJECT_BUSY = 17
    NO_SPACE_LEFT = 18
    OP_WOULD_BLOCK = 19
    IO_ERROR = 20
    OP_INTERRUPTED = 21
    OP_IN_PROGRESS = 22
    INTERNAL_ERROR = 23

    MUX_ERROR = 30
    NO_MEM = 31
    NOT_ENOUGH_DATA = 32
    DIR_NOT_EMPTY = 33

# See also: https://www.theiphonewiki.com/wiki/Models
MODELS = {
    "iPhone5,1": "iPhone 5",
    "iPhone5,2": "iPhone 5",
    "iPhone5,3": "iPhone 5c",
    "iPhone5,4": "iPhone 5c",
    "iPhone6,1": "iPhone 5s",
    "iPhone6,2": "iPhone 5s",
    "iPhone7,1": "iPhone 6 Plus",
    "iPhone7,2": "iPhone 6",
    "iPhone8,1": "iPhone 6s",
    "iPhone8,2": "iPhone 6s Plus",
    "iPhone8,4": "iPhone SE (1st generation)",
    "iPhone9,1": "iPhone 7",  # Global
    "iPhone9,2": "iPhone 7 Plus",  # Global
    "iPhone9,3": "iPhone 7",  # GSM
    "iPhone9,4": "iPhone 7 Plus",  # GSM
    "iPhone10,1": "iPhone 8",  # Global
    "iPhone10,2": "iPhone 8 Plus",  # Global
    "iPhone10,3": "iPhone X",  # Global
    "iPhone10,4": "iPhone 8",  # GSM
    "iPhone10,5": "iPhone 8 Plus",  # GSM
    "iPhone10,6": "iPhone X",  # GSM
    "iPhone11,2": "iPhone XS",
    "iPhone11,4": "iPhone XS Max",
    "iPhone11,6": "iPhone XS Max",
    "iPhone11,8": "iPhone XR",
    "iPhone12,1": "iPhone 11",
    "iPhone12,3": "iPhone 11 Pro",
    "iPhone12,5": "iPhone 11 Pro Max",
    "iPhone12,8": "iPhone SE (2nd generation)",
    "iPhone13,1": "iPhone 12 mini",
    "iPhone13,2": "iPhone 12",
    "iPhone13,3": "iPhone 12 Pro",
    "iPhone13,4": "iPhone 12 Pro Max",
    "iPhone14,2": "iPhone 13 Pro",
    "iPhone14,3": "iPhone 13 Pro Max",
    "iPhone14,4": "iPhone 13 mini",
    "iPhone14,5": "iPhone 13",

    "iPad11,1": "iPad mini (5th generation)",
    "iPad11,2": "iPad mini (5th generation)",
    "iPad5,2": "iPad mini 4",
    "iPad5,1": "iPad mini 4",
    "iPad4,7": "iPad mini 3",
    "iPad4,8": "iPad mini 3",
    "iPad4,9": "iPad mini 3",
    "iPad4,4": "iPad mini 2",
    "iPad4,5": "iPad mini 2",
    "iPad4,6": "iPad mini 2",
    "iPad2,5": "iPad mini",
    "iPad2,6": "iPad mini",
    "iPad2,7": "iPad mini",

    "iPad1,1": "iPad 1",
    "iPad2,1": "iPad (2rd generation)",
    "iPad2,2": "iPad (2rd generation)",
    "iPad2,3": "iPad (2rd generation)",
    "iPad2,4": "iPad (2rd generation)",
    "iPad3,1": "iPad (3rd generation)",
    "iPad3,2": "iPad (3rd generation)",
    "iPad3,3": "iPad (3rd generation)",
    "iPad3,4": "iPad (4rd generation)",
    "iPad3,5": "iPad (4rd generation)",
    "iPad3,6": "iPad (4rd generation)",
    "iPad6,11": "iPad (5th generation)",
    "iPad6,12": "iPad (5th generation)",
    "iPad7,5": "iPad (6th generation)",
    "iPad7,6": "iPad (6th generation)",
    "iPad7,11": "iPad (7th generation)",
    "iPad7,12": "iPad (7th generation)",
    "iPad11,6": "iPad (8th generation)",
    "iPad11,7": "iPad (8th generation)",

    "iPad13,1": "iPad Air (4th generation)",
    "iPad13,2": "iPad Air (4th generation)",
    "iPad11,3": "iPad Air (3th generation)",
    "iPad11,4": "iPad Air (3th generation)",
    "iPad5,3": "iPad Air 2",
    "iPad5,4": "iPad Air 2",
    "iPad4,1": "iPad Air",
    "iPad4,2": "iPad Air",
    "iPad4,3": "iPad Air",

    "iPad8,11": "iPad Pro (12.9-inch) (4th generation)",
    "iPad8,12": "iPad Pro (12.9-inch) (4th generation)",
    "iPad8,9": "iPad Pro (11-inch) (2nd generation)",
    "iPad8,10": "iPad Pro (11-inch) (2nd generation)",
    "iPad8,5": "iPad Pro (12.9-inch) (3rd generation)",
    "iPad8,6": "iPad Pro (12.9-inch) (3rd generation)",
    "iPad8,7": "iPad Pro (12.9-inch) (3rd generation)",
    "iPad8,8": "iPad Pro (12.9-inch) (3rd generation)",
    "iPad8,1": "iPad Pro (11-inch)",
    "iPad8,2": "iPad Pro (11-inch)",
    "iPad8,3": "iPad Pro (11-inch)",
    "iPad8,4": "iPad Pro (11-inch)",
    "iPad7,3": "iPad Pro (10.5-inch)",
    "iPad7,4": "iPad Pro (10.5-inch)",
    "iPad7,1": "iPad Pro (12.9-inch) (2nd generation)",
    "iPad7,2": "iPad Pro (12.9-inch) (2nd generation)",
    "iPad6,3": "iPad Pro (9.7-inch)",
    "iPad6,4": "iPad Pro (9.7-inch)",
    "iPad6,7": "iPad Pro (12.9-inch)",
    "iPad6,8": "iPad Pro (12.9-inch)",

    # simulator
    "i386": "iPhone Simulator",
    "x86_64": "iPhone Simulator",
}


class UsbmuxMessageType(str, enum.Enum):
    Attached = "Attached"
    Detached = "Detached"


class LockdownService(str, enum.Enum):
    # hdiutil mount /Applications/Xcode.app/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/14.0/DeveloperDiskImage.dmg
    # tree /Volumes/DeveloperDiskImage/Library/Lockdown
    MobileLockdown = 'com.apple.mobile.lockdown'
    
    CRASH_REPORT_MOVER_SERVICE = "com.apple.crashreportmover"
    CRASH_REPORT_COPY_MOBILE_SERVICE = "com.apple.crashreportcopymobile"

    # Ref: https://github.com/anonymous5l/iConsole/blob/master/wifiSync.go
    MobileWirelessLockdown = "com.apple.mobile.wireless_lockdown"
    InstallationProxy = "com.apple.mobile.installation_proxy"

    MobileScreenshotr = "com.apple.mobile.screenshotr"  # 截图服务
    MobileHouseArrest = "com.apple.mobile.house_arrest"  # 访问文件内的沙箱
    AFC = "com.apple.afc"  # 访问系统资源

    InstrumentsRemoteServer = "com.apple.instruments.remoteserver"
    InstrumentsRemoteServerSecure = "com.apple.instruments.remoteserver.DVTSecureSocketProxy"  # for iOS 14.0
    TestmanagerdLockdown = "com.apple.testmanagerd.lockdown"
    TestmanagerdLockdownSecure = "com.apple.testmanagerd.lockdown.secure"  # for iOS 14.0

    


class InstrumentsService(str, enum.Enum):
    DeviceInfo = 'com.apple.instruments.server.services.deviceinfo'
    ProcessControl = "com.apple.instruments.server.services.processcontrol"
    DeviceApplictionListing = "com.apple.instruments.server.services.device.applictionListing"
    GraphicsOpengl = "com.apple.instruments.server.services.graphics.opengl"  # 获取FPS
    Sysmontap = "com.apple.instruments.server.services.sysmontap"  # 获取性能数据用
    XcodeNetworkStatistics = 'com.apple.xcode.debug-gauge-data-providers.NetworkStatistics'  # 获取单进程网络数据
    Networking = 'com.apple.instruments.server.services.networking'  # 全局网络数据 instruments 用的就是这个
    MobileNotifications = 'com.apple.instruments.server.services.mobilenotifications'  # 监控应用状态


class UsbmuxReplyCode(int, enum.Enum):
    """
    Ref: https://github.com/libimobiledevice/usbmuxd/blob/master/src/usbmuxd-proto.h
    """
    OK = 0
    BadCommand = 1
    BadDevice = 2
    ConnectionRefused = 3
    BadVersion = 6


@dataclass
class StatResult:
    st_ifmt: str
    st_size: int
    st_blocks: int
    st_nlink: int
    st_ctime: datetime.datetime
    st_mtime: datetime.datetime
    st_linktarget: str = None

    def is_dir(self) -> bool:
        return self.st_ifmt == "S_IFDIR"
    
    def is_link(self) -> bool:
        return self.st_ifmt == "S_IFLNK"




if __name__ == "__main__":
    print(Color.wrap_text("Hello", Color.RED))
    print(AFC.GET_DEVINFO)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Feb 01 2021 14:09:21 by codeskyblue

Same as iproxy
"""

import colored
import simple_tornado
from logzero import logger
from tornado.ioloop import IOLoop
from tornado.iostream import IOStream, StreamClosedError
from tornado.tcpserver import TCPServer

from ._device import Device
from ._hexdump import hexdump
from ._safe_socket import PlistSocket
from .exceptions import MuxReplyError


class RelayTCPServer(TCPServer):
    def __init__(self, device: Device, device_port: int, debug: bool = False):
        """
        Args:
            port (int): forward to port
        """
        self.__device = device
        self.__port = device_port
        self.__debug = debug
        self.__names = {}
        super().__init__()

    # Override
    async def handle_stream(self, stream: IOStream, address):
        d = self.__device
        logger.debug("[%s] handle stream from: %s", d.udid, address)
        d._info = None # Force to refresh devId
        try:
            plconn = d.create_inner_connection(self.__port)
            asock = IOStream(plconn._sock)
            self._pipe_twoway(asock, stream, plconn)

            self.__names[stream] = address
        except MuxReplyError as e:
            logger.error("connect to device error: %s", e)
            stream.close()

    def _pipe_twoway(self, _in: IOStream, out: IOStream, plconn):
        io_loop = IOLoop.current()
        io_loop.add_callback(self._pipe_stream, _in, out, plconn)
        io_loop.add_callback(self._pipe_stream, out, _in, plconn)

    async def _pipe_stream(self, _in: IOStream, out: IOStream, plconn: PlistSocket):
        while not _in.closed():
            try:
                data = await _in.read_bytes(10240, partial=True)
                if self.__debug:
                    if self.__names.get(_in):
                        print("{} >>> device".format(self.__names[_in]))
                        print(colored.fg("green") + colored.bg("black"), end="", flush=True)
                    else:
                        print("device >>> {}".format(self.__names[out]))
                        print(colored.fg("orchid") + colored.bg("black"), end="", flush=True)
                    hexdump(data)
                    print(colored.attr("reset"), end="", flush=True)
                await out.write(data)
            except StreamClosedError as e:
                break

        # here may call twice
        _in.close() 
        out.close()
        plconn.close()
        self.__names.pop(_in, None)


def relay(d: Device, lport: int, rport: int, debug: bool = False):
    """
    relay tcp data from pc to device

    Args:
        lport: local port
        rport: remote port
    """
    simple_tornado.patch_for_windows()
    RelayTCPServer(device=d, device_port=rport, debug=debug).listen(lport)

    try:
        IOLoop.instance().start()
    except KeyboardInterrupt:
        IOLoop.instance().stop()



# coding: utf-8
# codeskyblue 2020/06/03
#

__all__ = ['SafeStreamSocket', 'PlistSocket']

import logging
import os
import plistlib
import socket
import ssl
import struct
import threading
import typing
import weakref
from typing import Any, Union

from ._proto import PROGRAM_NAME
from ._utils import set_socket_timeout
from .exceptions import *

logger = logging.getLogger(PROGRAM_NAME)

_n = [0]
_nlock = threading.Lock()

def get_uniq_id() -> int:
    with _nlock:
        _n[0] += 1
        return _n[0]


class SafeStreamSocket:
    def __init__(self, addr: Union[str, tuple, socket.socket,
                                   Any]):
        """
        Args:
            addr: can be /var/run/usbmuxd or (localhost, 27015)
        """
        self._id = get_uniq_id()
        self._sock = None
        self._name = None

        if isinstance(addr, socket.socket):
            self._sock = addr
        else:
            if isinstance(addr, str):
                if ':' in addr:
                    host, port = addr.split(":", 1)
                    addr = (host, int(port))
                    family = socket.AF_INET
                elif os.path.exists(addr):
                    family = socket.AF_UNIX
                else:
                    raise MuxError("socket unix:{} unable to connect".format(addr))
            else:
                family = socket.AF_INET
            self._sock = socket.socket(family, socket.SOCK_STREAM)
            self._sock.connect(addr)
        
        self._sock_gclist = [self._sock]

        def _cleanup(socks: typing.List[socket.socket]):
            _id = str(self.id)
            if self.name:
                _id = self.name + ":" + str(self.id)
            logger.debug("CLOSE(%s)", _id)
            for sock in socks:
                sock.close()

        self._finalizer = weakref.finalize(self, _cleanup, self._sock_gclist)
    
    def close(self):
        self._finalizer()
        
    @property
    def closed(self) -> bool:
        return not self._finalizer.alive

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name
    
    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    def get_socket(self) -> socket.socket:
        return self._sock

    def recv(self, bufsize: int = 4096) -> bytes:
        return self._sock.recv(bufsize)

    def recvall(self, size: int) -> bytearray:
        buf = bytearray()
        while len(buf) < size:
            chunk = self._sock.recv(size - len(buf))
            if not chunk:
                raise MuxError("socket connection broken")
            buf.extend(chunk)
        return buf

    def sendall(self, data: Union[bytes, bytearray]) -> int:
        with set_socket_timeout(self._sock, 10):
            return self._sock.sendall(data)

    def switch_to_ssl(self, pemfile):
        """ wrap socket to SSLSocket """
        # logger.debug("Switch to ssl")
        assert os.path.isfile(pemfile)
        self._dup_sock = self._sock.dup()
        self._sock_gclist.append(self._dup_sock)
        
        # https://docs.python.org/zh-cn/3/library/ssl.html#ssl.SSLContext
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        try:
            context.set_ciphers("ALL:@SECLEVEL=0") # fix md_too_weak error
        except ssl.SSLError:
            # ignore: no ciphers can be selected.
            pass
        context.load_cert_chain(pemfile, keyfile=pemfile)
        context.check_hostname = False
        ssock = context.wrap_socket(self._sock, server_hostname="iphone.localhost")
        
        self._sock = ssock

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class PlistSocket(SafeStreamSocket):
    def __init__(self, addr: str, tag: int = 0):
        super().__init__(addr)
        if isinstance(addr, PlistSocket):
            self._tag = addr._tag
            self._first = addr._first
        else:
            self._tag = tag
            self._first = True
        self.prepare()

    def prepare(self):
        pass

    def is_secure(self):
        return isinstance(self._sock, ssl.SSLSocket)

    def send_packet(self, payload: dict, message_type: int = 8):
        """
        Args:
            payload: required

            # The following args only used in the first request
            message_type: 8 (Plist)
            tag: int
        """
        #if self.is_secure():
        #    logger.debug(secure_text + " send: %s", payload)
        #else:
        logger.debug("SEND(%d): %s", self.id, payload)

        body_data = plistlib.dumps(payload)
        if self._first:  # first package
            length = 16 + len(body_data)
            header = struct.pack(
                "IIII", length, 1, message_type,
                self._tag)  # version: 1, request: 8(?), tag: 1(?)
        else:
            header = struct.pack(">I", len(body_data))
        self.sendall(header + body_data)

    def recv_packet(self, header_size=None) -> dict:
        if self._first or header_size == 16:  # first receive
            header = self.recvall(16)
            (length, version, resp, tag) = struct.unpack("IIII", header)
            length -= 16  # minus header length
            self._first = False
        else:
            header = self.recvall(4)
            (length, ) = struct.unpack(">I", header)

        body_data = self.recvall(length)
        payload = plistlib.loads(body_data)
        if 'PairRecordData' in payload:
            logger.debug("Recv pair record data ...")
        else:
            # if self.is_secure():
            #    logger.debug(secure_text + " recv" + Color.END + ": %s",
            #                 payload)
            # else:
            logger.debug("RECV(%d): %s", self.id, payload)
        return payload

    def send_recv_packet(self, payload: dict) -> dict:
        self.send_packet(payload)
        return self.recv_packet()


class PlistSocketProperty:
    def __init__(self, psock: PlistSocket):
        self._psock = psock
        self.prepare()
        self._finalizer = weakref.finalize(self, self.psock.close)
    
    @property
    def psock(self) -> PlistSocket:
        return self._psock
    
    def prepare(self):
        pass

    def send_packet(self, payload: dict, message_type: int = 8):
        return self.psock.send_packet(payload, message_type)
    
    def recv_packet(self, header_size=None) -> dict:
        return self.psock.recv_packet(header_size)
    
    def close(self):
        self._finalizer()
    
    @property
    def closed(self) -> bool:
        return not self._finalizer.alive
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue May 11 2021 16:30:17 by codeskyblue
"""

import enum
import io
import threading
import time
import typing
import uuid
from collections import defaultdict, namedtuple
from typing import Any, Iterator, Optional, Tuple, Union

from ._device import BaseDevice
from ._proto import *


class DataType(str, enum.Enum):
    SCREENSHOT = "screenshot"
    CPU = "cpu"
    MEMORY = "memory"
    NETWORK = "network"  # 流量
    FPS = "fps"
    PAGE = "page"
    GPU = "gpu"

CallbackType = typing.Callable[[DataType, dict], None]

class RunningProcess:
    """ acturally there is a better way to monitor process pid """
    PID_UPDATE_DURATION = 5.0

    def __init__(self, d: BaseDevice, bundle_id: str):
        self._ins = d.instruments
        self._bundle_id = bundle_id
        self._app_infos = list(d.installation.iter_installed(app_type=None))
        self._next_update_time = 0.0
        self._last_pid = None
        self._lock = threading.Lock()

    @property
    def bundle_id(self) -> str:
        return self._bundle_id

    def get_pid(self) -> Union[int, None]:
        """ return pid """
        with self._lock:
            if time.time() < self._next_update_time:
                return self._last_pid

            if self._last_pid and self._ins.is_running_pid(self._last_pid):
                self._next_update_time = time.time() + self.PID_UPDATE_DURATION
                return self._last_pid

            for pinfo in self._ins.app_process_list(self._app_infos):
                if pinfo['bundle_id'] == self._bundle_id:
                    self._last_pid = pinfo['pid']
                    self._next_update_time = time.time(
                    ) + self.PID_UPDATE_DURATION
                    # print(self._bundle_id, "pid:", self._last_pid)
                    return self._last_pid


class WaitGroup(object):
    """WaitGroup is like Go sync.WaitGroup.

    Without all the useful corner cases.
    """
    def __init__(self):
        self.count = 0
        self.cv = threading.Condition()

    def add(self, n):
        self.cv.acquire()
        self.count += n
        self.cv.release()

    def done(self):
        self.cv.acquire()
        self.count -= 1
        if self.count == 0:
            self.cv.notify_all()
        self.cv.release()

    # FIXME(ssx): here should quit when timeout, but maybe not
    def wait(self, timeout: Optional[float] = None):
        self.cv.acquire()
        while self.count > 0:
            self.cv.wait(timeout=timeout)
        self.cv.release()


def gen_stimestamp(seconds: Optional[float] = None) -> str:
    """ 生成专门用于tmq-service.taobao.org平台使用的timestampString """
    if seconds is None:
        seconds = time.time()
    return int(seconds * 1000)


def iter_fps(d: BaseDevice) -> Iterator[Any]:
    for data in d.instruments.iter_opengl_data():
        fps = data['CoreAnimationFramesPerSecond'] # fps from GPU
        # print("FPS:", fps)
        yield DataType.FPS, {"fps": fps, "time": time.time(), "value": fps}


def iter_gpu(d: BaseDevice) -> Iterator[Any]:
    for data in d.instruments.iter_opengl_data():
        device_utilization = data['Device Utilization %']  # Device Utilization
        tiler_utilization = data['Tiler Utilization %'] # Tiler Utilization
        renderer_utilization = data['Renderer Utilization %'] # Renderer Utilization
        yield DataType.GPU, {"device": device_utilization, "renderer": renderer_utilization,
                             "tiler": tiler_utilization, "time": time.time(), "value": device_utilization}


def iter_screenshot(d: BaseDevice) -> Iterator[Tuple[DataType, dict]]:
    for img in d.iter_screenshot():
        _time = time.time()
        img.thumbnail((500, 500))  # 缩小图片已方便保存
        
        # example of convert image to bytes
        # buf = io.BytesIO()
        # img.save(buf, format="JPEG")

        # turn image to URL
        yield DataType.SCREENSHOT, {"time": _time, "value": img}



ProcAttrs = namedtuple("ProcAttrs", SYSMON_PROC_ATTRS)

def _iter_complex_cpu_memory(d: BaseDevice,
                            rp: RunningProcess) -> Iterator[dict]:
    """
    content in iterator

    - {'type': 'system_cpu',
        'sys': -1.0,
        'total': 55.21212121212122,
        'user': -1.0}
    - {'type': 'process',
        'cpu_usage': 2.6393411792622925,
        'mem_anon': 54345728,
        'mem_rss': 130760704,
        'pid': 1344}
    """
    for info in d.instruments.iter_cpu_memory():
        pid = rp.get_pid()

        if info is None or len(info) != 2:
            continue
        sinfo, pinfolist = info
        if 'CPUCount' not in sinfo:
            sinfo, pinfolist = pinfolist, sinfo

        if 'CPUCount' not in sinfo:
            continue

        cpu_count = sinfo['CPUCount']

        sys_cpu_usage = sinfo['SystemCPUUsage']
        cpu_total_load = sys_cpu_usage['CPU_TotalLoad']
        cpu_user = sys_cpu_usage['CPU_UserLoad']
        cpu_sys = sys_cpu_usage['CPU_SystemLoad']

        if 'Processes' not in pinfolist:
            continue

        # 这里的total_cpu_usage加起来的累计值大概在0.5~5.0之间
        total_cpu_usage = 0.0
        for attrs in pinfolist['Processes'].values():
            pinfo = ProcAttrs(*attrs)
            if isinstance(pinfo.cpuUsage, float):  # maybe NSNull
                total_cpu_usage += pinfo.cpuUsage

        cpu_usage = 0.0
        attrs = pinfolist['Processes'].get(pid)
        if attrs is None:  # process is not running
            # continue
            # print('process not launched')
            pass
        else:
            assert len(attrs) == len(SYSMON_PROC_ATTRS)
            # print(ProcAttrs, attrs)
            pinfo = ProcAttrs(*attrs)
            cpu_usage = pinfo.cpuUsage
        # next_list_process_time = time.time() + next_timeout
        # cpu_usage, rss, mem_anon, pid = pinfo

        # 很诡异的计算方法，不过也就这种方法计算出来的CPU看起来正常一点
        # 计算后的cpuUsage范围 [0, 100]
        # cpu_total_load /= cpu_count
        # cpu_usage *= cpu_total_load
        # if total_cpu_usage > 0:
        #     cpu_usage /= total_cpu_usage

        # print("cpuUsage: {}, total: {}".format(cpu_usage, total_cpu_usage))
        # print("memory: {} MB".format(pinfo.physFootprint / 1024 / 1024))
        yield dict(
            type="process",
            pid=pid,
            phys_memory=pinfo.physFootprint,  # 物理内存
            phys_memory_string="{:.1f} MiB".format(pinfo.physFootprint / 1024 /
                                                   1024),
            vss=pinfo.memVirtualSize,
            rss=pinfo.memResidentSize,
            anon=pinfo.memAnon,  # 匿名内存? 这个是啥
            cpu_count=cpu_count,
            cpu_usage=cpu_usage,  # 理论上最高 100.0 (这里是除以过cpuCount的)
            sys_cpu_usage=cpu_total_load,
            attr_cpuUsage=pinfo.cpuUsage,
            attr_cpuTotal=cpu_total_load,
            attr_ctxSwitch=pinfo.ctxSwitch,
            attr_intWakeups=pinfo.intWakeups,
            attr_systemInfo=sys_cpu_usage)


def iter_cpu_memory(d: BaseDevice, rp: RunningProcess) -> Iterator[Any]:
    for minfo in _iter_complex_cpu_memory(d, rp):  # d.iter_cpu_mem(bundle_id):
        yield DataType.CPU, {
            "timestamp": gen_stimestamp(),
            "pid": minfo['pid'],
            "value": minfo['cpu_usage'],  # max 100.0?, maybe not
            "sys_value": minfo['sys_cpu_usage'],
            "count": minfo['cpu_count']
        }
        yield DataType.MEMORY, {
            "pid": minfo['pid'],
            "timestamp": gen_stimestamp(),
            "value": minfo['phys_memory'] / 1024 / 1024,  # MB
        }


def set_interval(it: Iterator[Any], interval: float):
    while True:
        start = time.time()
        data = next(it)
        yield data
        wait = max(0, interval - (time.time() - start))
        time.sleep(wait)


def iter_network_flow(d: BaseDevice, rp: RunningProcess) -> Iterator[Any]:
    n = 0
    for nstat in d.instruments.iter_network():
        # if n < 2:
        #     n += 1
        #     continue
        yield DataType.NETWORK, {
            "timestamp": gen_stimestamp(),
            "downFlow": (nstat['rx.bytes'] or 0) / 1024,
            "upFlow": (nstat['tx.bytes'] or 0) / 1024
        }


def append_data(wg: WaitGroup, stop_event: threading.Event,
                idata: Iterator[Any], callback: CallbackType, filters: list):
    for _type, data in idata:
        assert isinstance(data, dict)
        assert isinstance(_type, DataType)

        if stop_event.is_set():
            wg.done()
            break

        if isinstance(data, dict) and "time" in data:
            stimestamp = gen_stimestamp(data.pop('time'))
            data.update({"timestamp": stimestamp})
        # result[_type].append(data)
        if _type in filters:
            callback(_type, data)
        # print(_type, data)

    stop_event.set()  # 当有一个中断，其他的全部中断，让错误暴露出来


class Performance():
    # PROMPT_TITLE = "tidevice performance"

    def __init__(self, d: BaseDevice, perfs: typing.List[DataType] = []):
        self._d = d
        self._bundle_id = None
        self._stop_event = threading.Event()
        self._wg = WaitGroup()
        self._started = False
        self._result = defaultdict(list)
        self._perfs = perfs

        # the callback function accepts all the data
        self._callback = None

    def start(self, bundle_id: str, callback: CallbackType = None):
        if not callback:
            # 默认不输出屏幕的截图（暂时没想好怎么处理）
            callback = lambda _type, data: print(_type.value, data, flush=True) if _type != DataType.SCREENSHOT and _type in self._perfs else None
        self._rp = RunningProcess(self._d, bundle_id)
        self._thread_start(callback)

    def _thread_start(self, callback: CallbackType):
        iters = []
        if DataType.CPU in self._perfs or DataType.MEMORY in self._perfs:
            iters.append(iter_cpu_memory(self._d, self._rp))
        if DataType.FPS in self._perfs:
            iters.append(iter_fps(self._d))
        if DataType.GPU in self._perfs:
            iters.append(iter_gpu(self._d))
        if DataType.SCREENSHOT in self._perfs:
            iters.append(set_interval(iter_screenshot(self._d), 1.0))
        if DataType.NETWORK in self._perfs:
            iters.append(iter_network_flow(self._d, self._rp))
        for it in (iters): # yapf: disable
            self._wg.add(1)
            threading.Thread(name="perf",
                             target=append_data,
                             args=(self._wg, self._stop_event, it,
                                   callback,self._perfs),
                             daemon=True).start()

    def stop(self): # -> PerfReport:
        self._stop_event.set()
        print("Stopped")
        # memory and fps will take at least 1 second to catch _stop_event
        # to make function run faster, we not using self._wg.wait(..) here
        # > self._wg.wait(timeout=3.0) # wait all stopped
        # > self._started = False

    def wait(self, timeout: float):
        return self._wg.wait(timeout=timeout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Jan 04 2021 17:59:30 by codeskyblue
"""

from ._device import BaseDevice as Device
from ._usbmux import Usbmux, ConnectionType
from ._perf import Performance, DataType
from .exceptions import *


# coding: utf-8

import sys
import setuptools

#if sys.version_info[:2] <= (3, 6):
#    sys.exit("\n*** tidevice requires Python version 3.7+")

setuptools.setup(
    setup_requires=['pbr'], pbr=True, python_requires=">=3.6")



# coding: utf-8
# created: codeskyblue 2020/06/19
#

import re
import shutil
import sys
import tempfile
import typing
import zipfile
from pprint import pprint
from typing import Union

from . import bplist, plistlib2
from ._compat import cache
from .exceptions import IPAError


class IPAReader(zipfile.ZipFile):
    def get_infoplist_zipinfo(self) -> zipfile.ZipInfo:
        re_infoplist = re.compile(r"^Payload/[^/]+\.app/Info.plist$")
        for zipinfo in self.filelist:
            if re_infoplist.match(zipinfo.filename):
                return zipinfo
        raise IPAError("Info.plist not found")

    def get_mobileprovision_zipinfo(self) -> zipfile.ZipInfo:
        re_provision_file = re.compile(r"^Payload/[^/]+\.app/embedded.mobileprovision$")
        for zipinfo in self.filelist:
            if re_provision_file.match(zipinfo.filename):
                return zipinfo
        raise IPAError("embedded.mobileprovision not found")

    def get_mobileprovision(self) -> dict:
        """
        mobileprovision usally contains keys
            AppIDName, ApplicationIdentifierPrefix, CreationDate,
            Platform, IsXcodeManaged, DeveloperCertificates,
            Entitlements, ExpirationDate, Name, ProvisionedDevices,
            TeamIdentifier, TeamName,
            TimeToLive, UUID, Version
        """
        provision_xml_rx = re.compile(br'<\?xml.+</plist>', re.DOTALL)
        content = self.read(self.get_mobileprovision_zipinfo())
        match = provision_xml_rx.search(content)
        if match:
            xml_content = match.group()
            data = plistlib2.loads(xml_content)
            return data
        else:
            raise IPAError('unable to parse embedded.mobileprovision file')

    @cache
    def get_infoplist(self) -> dict:
        finfo = self.get_infoplist_zipinfo()
        with self.open(finfo, 'r') as fp:
            if sys.version_info[:2] <= (3, 6):
                # for compatiable with py3.6
                with tempfile.TemporaryFile() as tmpf:
                    shutil.copyfileobj(fp, tmpf)
                    tmpf.seek(0)
                    return bplist.load(tmpf)
            return bplist.load(fp)

    def get_bundle_id(self) -> str:
        """ return CFBundleIdentifier """
        return self.get_infoplist()['CFBundleIdentifier']
    
    def get_short_version(self) -> str:
        return self.get_infoplist().get('CFBundleShortVersionString', "")

    def dump_info(self, all: bool = False):
        data = self.get_infoplist()
        print("BundleID:", data['CFBundleIdentifier'])
        print("ShortVersion:", data['CFBundleShortVersionString'])
        if all:
            m = self.get_mobileprovision()
            pprint(m['ProvisionedDevices'])


def parse_bundle_id(fpath: str) -> str:
    with open(fpath, "rb") as f:
        return IPAReader(f).get_bundle_id()


def main():
    # ir = IPAReader("../??.ipa")
    # print(ir.get_bundle_id())
    url = "???"  # FIXME(ssx): here need a public ipa package
    import httpio
    with httpio.open(url, block_size=-1) as fp:
        ir = IPAReader(fp)
        print(ir.get_bundle_id())


if __name__ == "__main__":
    main()
